// Solution: forblast   Project: fastaprep
#include "fastaprep.h"

string  inpath = def_inpath;    // input fasta file path
string  stapath = def_stapath;  // status file path
string  staext = def_staext;	// status file ext
string	dupath = def_dupath;	// duplicates file path
string	dupext = def_dupext;	// duplicates file ext
string  outpath = def_outpath;  // output fasta file path
string  outext = def_outext;    // output file ext
int		order = def_order;		// protein ordering: -1=size> 1=size< 2=lexico 0=unsorted
char    buf[buflen];
clock_t t0, t1, t2;
vector <string> species;		// array of species names
vector <string> fasta;			// array of FASTA file names
vector <string> allspecies;		// array of all species names
vector <string> allfasta;		// array of all FASTA file names
int m;							// number of active species
int n;							// number of all species
int copy_mode = def_cmode;
int status_mode = def_smode;
int write_mode = def_wmode;

int main (int argc, char** argv)
{
	string mypgm(argv[0]);
	try
	{
		string::size_type pos = mypgm.find_last_of(slash);
		if (pos != string::npos)
			mypgm.erase(0, pos + 1);
		pos = mypgm.find_last_of(".");
		if (pos != string::npos)
			mypgm.erase(pos);
	}
	catch (...) {}

	if (argc <= 1 || argc > 1 && (argv[1][0] == '?' ||
		(argv[1][0] == '-' || argv[1][0] == '/') &&
		(argv[1][1] == 'H' || argv[1][1] == 'h' || argv[1][1] == '?')))
    {
        showHelp(mypgm);
        return 0;
    }

    t0 = clock();

	// copy arguments to the log
	cout << "\n" << mypgm << " v" << version << ":";
    for (int i = 1; i < argc; i++)
        cout << " " << argv[i];
    cout << "\n" << endl;
        
	ifstream lst;
	string listname;
	string allname;
	string::size_type pos, off;

    // parse arguments
    for (int i = 1; i < argc; i++)
    {
		if (argv[i][0] == '-')
		{
			char option = argv[i][1];
			switch (option)
			{
            case 'C':
            case 'c':
                if (++i < argc)
                    copy_mode = argv[i][0] == '0' ? 0 : 1;
                else
                {
					cout << "Missing copy_mode value" << endl;
					return 1;
				}
                break;
			case 'D':
			case 'd':
				if (++i < argc)
					dupath.assign(argv[i]);
				else
				{
					cout << "Missing path of protein duplicates" << endl;
					return 1;
				}
				break;
            case 'E':
            case 'e':
                if (++i < argc)
                    dupext.assign(argv[i]);
                else
                {
                    cout << "Missing extension of protein duplicate files" << endl;
                    return 1;
                }
                break;
            case '?':
			case 'H':
			case 'h':
				showHelp(mypgm);
				return 0;
				break;
			case 'I':
			case 'i':
				if (++i < argc)
					inpath.assign(argv[i]);
				else
				{
					cout << "Missing input path" << endl;
					return 1;
				}
				break;
			case 'O':
			case 'o':
				if (++i < argc)
					outpath.assign(argv[i]);
				else
				{
					cout << "Missing output path" << endl;
					return 1;
				}
				break;
			case 'P':
			case 'p':
				if (++i < argc)
					outext.assign(argv[i]);
				else
				{
					cout << "Missing output file extension" << endl;
					return 1;
				}
				break;
			case 'R':
			case 'r':
				if (++i < argc)
				{
					string str_order(argv[i]);
					istringstream iss(str_order);
					iss >> order;
				}
				else
				{
					cout << "Missing value of ordering" << endl;
					return 1;
				}
				break;
			case 'S':
			case 's':
				if (++i < argc)
					stapath.assign(argv[i]);
				else
				{
					cout << "Missing status path" << endl;
					return 1;
				}
				break;
            case 'T':
            case 't':
                if (++i < argc)
                    staext.assign(argv[i]);
                else
                {
                    cout << "Missing status file extension" << endl;
                    return 1;
                }
                break;
            case 'U':
            case 'u':
                if (++i < argc)
                    status_mode = argv[i][0] == '0' ? 0 : 1;
                else
                {
                    cout << "Missing status_mode value" << endl;
                    return 1;
                }
                break;
            case 'W':
            case 'w':
                if (++i < argc)
                    write_mode = argv[i][0] == '0' ? 0 : 1;
                else
                {
                    cout << "Missing write_mode value" << endl;
                    return 1;
                }
                break;
            default:
				cout << "Unknown option " << option << endl;
				return 1;
				break;
			}
		}
		else if (listname.empty())
			listname.assign(argv[i]);
		else if (allname.empty())
			allname.assign(argv[i]);
		else
        {
			string extra(argv[i]);
			cout << "Extra argument " << extra << " ignored";
        }
    }

	// check that both lists are specified
	if (listname.empty() || allname.empty())
	{
		cout << "Two lists of species must be specified" << endl;
		return 1;
	}

	// refine parameters
	if (!inpath.empty() && *inpath.rbegin() != slash && *inpath.rbegin() != slash2)
		inpath.append(1, slash);
	if (!outpath.empty() && *outpath.rbegin() != slash && *outpath.rbegin() != slash2)
		outpath.append(1, slash);
    if (!outext.empty() && outext[0] != '.')
        outext = "." + outext;
	if (!stapath.empty() && *stapath.rbegin() != slash && *stapath.rbegin() != slash2)
		stapath.push_back(slash);
	if (!staext.empty() && staext[0] != '.')
		staext = "." + staext;
	if (!dupath.empty() && *dupath.rbegin() != slash && *dupath.rbegin() != slash2)
		dupath.push_back(slash);
	if (!dupext.empty() && dupext[0] != '.')
		dupext = "." + dupext;

	// read all_species_list file
    lst.open(allname.c_str(), ios_base::binary);
    if (lst.bad()) 
    {
        cout << "Bad or missing list of all species " << allname << endl;
        return 2;
    }

    int lineno = 1;
	for (; !lst.eof(); lineno++)
    {
        lst.getline(buf, buflen);
        if (buf[0] != 0 && lst.fail()) 
        {
            cout << "Failed at line " << lineno << endl;
            return 2;
        }

		string alias;
		string filename;
        string line(buf);
        if (!line.empty() && *line.rbegin() == '\r')
            line.erase(line.size() - 1);

        if (lineno == 1 || line.empty())    // skip first and last lines
            continue;
        if (comment.find(line[0]) != string::npos)
            continue;

		// parse alias (species name)
        pos = line.find_first_of(tab);
		if (pos == string::npos)
		{
			cout << "no 2nd field in line " << lineno << endl;
			return 3;
		}
		else
			alias = line.substr(0, pos);

		// parse filename
		off = line.find_first_not_of(tab, pos);
		if (off == string::npos)
		{
			cout << "no file name in line " << lineno << endl;
			return 3;
		}
		else
		{
			pos = line.find_first_of(tab, off);
			if (pos == string::npos)
				filename = line.substr(off);
			else
				filename = line.substr(off, pos - off);
		}

		allspecies.push_back(alias);
		allfasta.push_back(filename);
    }
	lst.clear(); lst.close();
	n = (int)allspecies.size();

	string flags(n, ' ');

	// read active species list file
	lst.open(listname.c_str(), ios_base::binary);
    if (lst.bad()) 
    {
        cout << "Bad or missing list of active species, " << allname << endl;
        return 2;
    }

    lineno = 1;
	for (; !lst.eof(); lineno++)
    {
        lst.getline(buf, buflen);
        if (buf[0] != 0 && lst.fail()) 
        {
            cout << "Failed at line " << lineno << endl;
            return 2;
        }

		string alias;
		string filename;
        string line(buf);
        if (!line.empty() && *line.rbegin() == '\r')
            line.erase(line.size() - 1);

        if (lineno == 1 || line.empty())
            continue;
        if (comment.find(line[0]) != string::npos)
            continue;

		// parse alias (species name)
        pos = line.find_first_of(tab);
		if (pos == string::npos)
		{
			cout << "no 2nd field in line " << lineno << endl;
			return 3;
		}
		else
			alias = line.substr(0, pos);

		// parse filename
		off = line.find_first_not_of(tab, pos);
		if (off == string::npos)
		{
			cout << "no file name in line " << lineno << endl;
			return 3;
		}
		else
		{
			pos = line.find_first_of(tab, off);
			if (pos == string::npos)
				filename = line.substr(off);
			else
				filename = line.substr(off, pos - off);
		}

		int k;
		for (k = 0; k < n; k++)
			if (allspecies[k] == alias && allfasta[k] == filename)
				break;
		if (k < n)
		{
			species.push_back(alias);
			fasta.push_back(filename);
			flags[k] = dot;
		}
		else
		{
			cout << "Species " << alias << " disagrees with master list" << endl;
			return 4;
		}
    }
	lst.clear(); lst.close();
	m = (int)species.size();

	cout << "Number of species to activate: " << m << " of " << n << endl;

    // process species at turn
    int doneok = 0;
    int failed = 0;
    for (int k = 0; k < n; k++)
    {
        t1 = clock();
        if (convert(k, flags)) doneok++;
        else failed++;
    }

    t2 = clock();
    clock_t mm = (t2 - t0)/CLOCKS_PER_SEC/60;
    cout << "\nLines:" << lineno-1 << "  Species_OK:" << doneok 
        << "  Errors:" << failed << "  Time:" << mm << "m" << endl;
}

bool convert(int k, string & flags)
{
    int tdups = 0;      // total of copies
    int maxdups = 0;    // max # of copies per group
    int pdups = 0;      // proteins with at least one copy
	string alias = allspecies[k];
	string fastaname = allfasta[k];
    vector <Protein> proteins;

    cout << left << setw(maxw) << alias+"..." << right;

    // read entire input file
    string filename = inpath + fastaname;
    ifstream ifa(filename.c_str(), ios_base::binary);
    if (ifa.bad())
    {
        cout << "can't open input file " << filename << endl;
        return false;
    }

    int lineno = 0;
	int offset = 0;
	string::size_type pos;
    Protein protein;
	int protno = 0;		// serial number of the protein

    while (!ifa.eof())
    {
		offset = (int)ifa.tellg();
        ifa.getline(buf, buflen);
        if (buf[0] != 0 && ifa.fail())
        {
            cout << "failed at line " << lineno+1 << endl;
            return false;
        }
        lineno++;

        string line(buf);
        if (!line.empty() && *line.rbegin() == '\r')
            line.erase(line.size() - 1);
        if (line.empty())
        {
            if (!ifa.eof())
            {
                cout << "empty line " << lineno << endl;
                return false;
            }
            if (!protein.sequence.empty())
            {
				protein.num = protno++;
                proteins.push_back(protein);
                protein.id.clear();
                protein.sequence.clear();
                break;
            }
        }
        else if (line[0] == hdrstart)		// parse header
        {
            if (!protein.sequence.empty())
            {
				protein.num = protno++;
                proteins.push_back(protein);
                protein.id.clear();
                protein.sequence.clear();
            }
            else if (!protein.id.empty())
            {
                cout << "no sequence at line " << lineno << endl;
                return false;
            }

			protein.offset = offset;
            pos = line.find_first_of(delim);
            if (pos != string::npos)
                line.erase(pos);
            protein.id = line.substr(1);
			protein.lineno = lineno;
        }
        else								// append sequence
        {
            pos = line.find_first_of(eol);
            if (pos != string::npos)
                line.erase(pos);
            protein.sequence.append(line);
        }
    }
	if (!protein.id.empty() && !protein.sequence.empty())
	{
		protein.num = protno++;
        proteins.push_back(protein);
	}

    // identify and count duplicates
    if (copy_mode == 1)
    {
        Seqs seqs;				// map of already met sequences

        for (int i = 0; i < (int)proteins.size(); i++)
        {
            Protein& proti = proteins[i];
            string& seq = proti.sequence;
            pair <SeqsIt, bool> res = seqs.insert(SeqPair(seq, i));

            if (!res.second)
            {
                proti.isdup = true;
                tdups++;

                int j = res.first->second;
                Protein& protj = proteins[j];

                if (protj.copy.empty())
                    pdups++;

                protj.copy.push_back(i);
                if ((int)protj.copy.size() > maxdups)
                    maxdups = (int)protj.copy.size();
            }

        }

        cout << " p=" << left << setw(6) << proteins.size() << " dup=" << setw(5)
            << tdups << " max=" << setw(2) << maxdups << " grp=" << setw(5)
            << pdups << right;
    }

#ifdef FASTAPREP_VERIFY
    vector <Protein> old_proteins(proteins);
#endif	// FASTAPREP_VERIFY
    if (order != 0)
    {
        sort(proteins.begin(), proteins.end(), greaterProtein);

        // fill index of initial protein numbers
        vector <int> index(proteins.size());
        for (int k = 0; k < (int)proteins.size(); k++)
        {
            Protein& prot = proteins[k];
            index[prot.num] = k;
        }

        // renumber copies according to new order
        if (copy_mode == 1)
        {
            for (int k = 0; k < (int)proteins.size(); k++)
            {
                Protein& prot = proteins[k];
                for (int i = 0; i < (int)prot.copy.size(); i++)
                    prot.copy[i] = index[prot.copy[i]];
            }
        }
    }

    // create duplicate file
    ofstream odu;
    if (copy_mode == 1)
    {
        filename = dupath + alias + dupext;
        PERMISSIONS odu.open(filename.c_str(), ios_base::binary);
        if (odu.bad())
        {
            cout << "can't create duplicate file " << filename << endl;
            return false;
        }
    }

    // create status file
    ofstream ost;
    if (status_mode == 1)
    {
        filename = stapath + alias + staext;
        PERMISSIONS ost.open(filename.c_str(), ios_base::binary);
        if (ost.bad())
        {
            cout << "can't create status file " << filename << endl;
            return false;
        }
    }

    // create output fasta file
    ofstream ofa;
    if (write_mode == 1)
    {
        filename = outpath + alias + outext;
        PERMISSIONS ofa.open(filename.c_str(), ios_base::binary);
        if (ofa.bad())
        {
            cout << "can't create output FASTA file " << filename << endl;
            return false;
        }
    }

    // write needed result files
    int n = (int)allspecies.size();		// number of species
    string grid (n, dot);					// for active species
	if (flags[k] != dot) grid = flags;		// for inactive species
	int stano = 1;
    int fasno = 0;

    for (int i = 0; i < (int)proteins.size(); i++)
    {
        Protein& proti = proteins[i];

		if (copy_mode == 1 && proti.isdup) continue;			// don't write all files

	#ifdef FASTAPREP_VERIFY
		// verify consistency
		int n = proti.num;
		Protein& oprotn = old_proteins[n];
		int dis = 0;
		if (proti.id != oprotn.id)									dis = 1;
		else if (proti.sequence.size() != oprotn.sequence.size())	dis = 2;
		else if (proti.sequence != oprotn.sequence)					dis = 3;
		else if (proti.offset != oprotn.offset)						dis = 4;
		else if (proti.lineno != oprotn.lineno)						dis = 5;
		else if (proti.copy.size() != oprotn.copy.size())			dis = 6;
		else if (proti.isdup != oprotn.isdup)						dis = 7;
		if (dis)
		{
			cout << " discrepancy " << dis << " in status #" << stano 
				<< " and fasta #" << oprotn.lineno << " lines";
		}
	#endif	// FASTAPREP_VERIFY

        if (write_mode == 1)                                    // write output fasta file
        {
            ofa << hdrstart << proti.id << endl;
            string::size_type len = proti.sequence.size();
            string::size_type pos = 0;
            for (; pos + fastalen <= len; pos += fastalen)
                ofa << proti.sequence.substr(pos, fastalen) << endl;
            if (pos < len)
                ofa << proti.sequence.substr(pos) << endl;
            fasno++;
        }

        if (status_mode == 1)
        {
            ost << setw(strw) << proti.lineno << " " << setw(numw) << proti.num
                << " " << setw(lenw) << proti.sequence.size() << " " << grid << endl;
            stano++;
        }

		if (copy_mode == 0 || proti.copy.empty()) continue;	    // don't write a file of dups

		odu << proti.id << tab << proti.copy.size();
        for (int j = 0; j < (int)proti.copy.size(); j++)
        {
            Protein& protj = proteins[proti.copy[j]];
            odu << tab << protj.id;
        }
        odu << endl;
    }

	if (status_mode == 1)   ost.close();
	if (copy_mode == 1)     odu.close();
    if (write_mode == 1)    ofa.close();

    if (write_mode == 1)
        cout << setw(6) << fasno << " of " << setw(6) << proteins.size() << " proteins written in ";
    t2 = clock();
    clock_t sec = (t2 - t1)/CLOCKS_PER_SEC;
    cout << setw(3) << sec << "s" << endl;
    return true;
}

bool greaterProtein(const Protein& p1, const Protein& p2)
{
	if (order == 2) return p1.id < p2.id;

	else if (p1.sequence.size() == p2.sequence.size())
	{
		if (p1.sequence == p2.sequence)
		{
			return p1.id < p2.id;
		}
		return p1.sequence < p2.sequence;
	}

	else if (order == -1)
		return p1.sequence.size() > p2.sequence.size();
	else /*if (order == 1)*/
		return p1.sequence.size() < p2.sequence.size();
}

void showHelp(const string& program)
{
	cout << "\nUsage (v" << version << "): " << program << " [options] active_species all_species\n"
         << "\n"
         << "Options (case insensitive):\n"
         << "-c {0|1}  : 1=eliminate protein copies, 0=no [" << def_cmode << "]\n"
		 << "-d string : directory for protein duplicate files [" << def_dupath << "]\n" 
		 << "-e string : extension of protein duplicate files [" << def_dupext << "]\n"
         << "-h -?     : show this help information\n"
         << "-i string : directory of input FASTA files [" << def_inpath << "]\n"
         << "-o string : directory for output FASTA files [" << def_outpath << "]\n"
         << "-p string : extension of output files [" << def_outext << "]\n"
         << "-r number : -1=size> 1=size< 2=lexico 0=unsorted [" << def_order << "]\n"
         << "-s string : directory for status files [" << def_stapath << "]\n"
         << "-t string : extension of status files [" << def_staext << "]\n"
         << "-u {0|1}  : 1=update status files, 0=no [" << def_smode << "]\n"
         << "-w {0|1}  : 1=write output FASTA, 0=no [" << def_wmode << "]\n"
         << "\n"
		 << "active_species: list of new or modified species\n"
		 << "all_species: full list of species in process\n"
		 << "line of species list: species_name\\tFASTA_file_name\\t...\n"
         << "\n"
         << "Example: " << program << " -w 1 -r 0 species.txt species.txt\n" 
		 << endl;
}
