// Solution: forblast   Project: fastaprep
#ifndef FASTAPREP_H
#define FASTAPREP_H

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <time.h>

//#define FASTAPREP_VERIFY

using namespace std;

#if defined(WIN32) || defined(_WIN32) || defined(_WIN64)
#define isFile(a)	_access((a), 0)		// only check existense
#define PERMISSIONS			// no special permission in Windows
const char slash    = '\\';
#elif defined(__unix__)
#define isFile(a)	access((a), F_OK)
#include <sys/types.h>
#include <sys/stat.h>
#define PERMISSIONS	umask(002);		// -> rw-rw-r--
const char slash    = '/';
#endif

typedef map <string, int> Seqs;
typedef Seqs::iterator SeqsIt;
typedef pair <string, int> SeqPair;

const string    version     = "1.10";
const char      hdrstart    = '>';
const char      tab         = '\t';
const char      dot         = '.';
const char      slash2      = '/';
const string    comment     = ";#";
const string    blank       = " \t\n\r";
const string    delim       = " |\t\n\r";
const string	eol			= "\n\r";
const string    def_inpath	= "protein";
const string    def_outpath = "fasta";
const string    def_outext  = ".fas";
const string	def_stapath = "status";
const string	def_staext	= "";
const string	def_dupath	= "dup";
const string	def_dupext	= ".tsv";
const int		def_order	= -1;	// -1=size> 1=size< 2=lexico 0=unsorted
const int       def_cmode   = 1;    // copy_mode:   1=eliminate duplicates, 0=no
const int       def_smode   = 1;    // status_mode: 1=write status file(s), 0=no
const int       def_wmode   = 0;    // write mode:  1=write deduplicated fasta, 0=no
const int       buflen      = 65536;
const int       maxw        = 36;       // max width of species name
const int		strw		= 7;		// max width of fasta line number
const int       numw        = 6;        // max width of protein number
const int		lenw		= 5;		// max width of length
const int       fastalen    = 80;       // max length of FASTA line at output

struct Status
{
	char str[strw];		// line number of the protein header
	char blank1;		// blank
	char num[numw];		// protein serial number as right aligned decimal 
    char blank2;        // blank
	char len[lenw];		// protein length as right aligned decimal
    char blank3;        // blank
    char flag[1];       // [num_species]
};

struct Protein
{
    string  id;
    string  sequence;
	int		offset;		// position of '>'
	int		num;		// protein serial number
	int		lineno;		// line number of the protein header
	vector <int> copy;	// protein numbers of copies of this sequence
    bool	isdup;		// true: it is a copy of other sequence 
    Protein()	{ isdup = false; }
};

bool convert(int k, string & flags);
bool greaterProtein(const Protein& p1, const Protein& p2);
void showHelp(const string& program);
#endif