// Solution: forblast   Project: posthit
#include "posthit.h"

string	inpath		= unspec;	// input path
string	inmid		= unspec;	// interspecies connector
string	inext		= unspec;	// input extension
string	dupath		= unspec;	// file path of duplicates
string	dupext		= unspec;	// extension of duplicates
string	genepath	= unspec;	// path of gene tables
string	geneext		= unspec;	// extension of gene tables
string	outpath		= unspec;	// output path
string	outext		= unspec;	// output extension
string	logfile;
string	pfield = def_pfield;	// name of protein column in gene tables
string	gfield = def_gfield;	// name of gene column in gene tables
string  rfield = def_rfield;	// name of region column in gene tables
string  sfield = def_sfield;	// name of starts column in gene tables
int		usedup = def_usedup;		// expand with duplicates: 1=yes, 0=no
int		aggregate = def_aggregate;	// aggregate proteins->genes: 1=yes, 0=no
int		symmetry = def_symmetry;	// symmetrize the matrices: 1=yes, 0=no
int		useweight = def_useweight;	// calculate and use weight: 1=yes, 0=no
int		verbose = def_verbose;	// verbose output to console: 1=yes, 0=no
int		maxhomo = def_maxhomo;	// max number of best homologs, 0=unlimited
int		addrank = def_addrank;	// add rank field: 0=no, 1=forward, 2=both
float   ratio = def_ratio;      // ratio cutoff: 0=normal rank
string	hsortmode = def_hsortmode;	// horizontal sort mode: n i p s w e
string	vsortmode = def_vsortmode;	// vertical sort mode: n i p s w e
char    buf[buflen];
int     mpisize = 1;
int     myrank = 0;
int     rankw = 1;
double  t0, t1, t2;
bool    is_subdir = true;       // hits are distributed over subdirectories
bool    is_name1 = true;        // hit files are named with spec1 
bool    is_name2 = true;        // hit files are named with spec2
vector <Species> species;       // full array of species
int     nsp;                    // total number of species
vector <Protein> proteins;		// full array of proteins
Dict	prodict;				// united protein dictionary: id->protno
vector <Row> pmatrix;			// protein*2 matrix
vector <Gene> genes;			// full array of genes
Dict	gendict;				// united gene dictionary: id->geneno
vector <Row> gmatrix;			// gene*2 matrix
Index	gmindex;				// index of gene matrix
vector <Column> rmatrix;        // backward rank matrix
Index   rmindex;                // index of backward rank matrix
int		nprots = 0;				// total number of proteins
int		ngenes = 0;				// total number of genes
int		nambig = 0;				// total number of protein/gene ambiguities
int		nunique = 0;			// total number of unique proteins
int		nduplicated = 0;		// total number of duplicated protein groups
int		nunknown = 0;			// total number of unknown proteins (lack in gene table)
int		total_rows = 0;			// total number of rows/columns in the matrix
int		expand_rows = 0;		//		the same for expanded matrix
int		gene_rows = 0;			//		the same for aggregated matrix
long long total_cells = 0;		// total number of nonempty cells in the matrix
long long expand_cells = 0;		//		the same for expanded matrix
long long gene_cells = 0;		//		the same for aggregated matrix
ofstream det;					// detailed log
ostringstream plog;				// for pre-logging prior to the log creation
int     current_ino;            // shared by outputResult and sortCells

int main(int argc, char** argv)
{
	string mypgm(argv[0]);
	try
	{
		string::size_type pos = mypgm.find_last_of(slash);
		if (pos != string::npos)
			mypgm.erase(0, pos + 1);
		pos = mypgm.find_last_of(".");
		if (pos != string::npos)
			mypgm.erase(pos);
	}
	catch (...) {}

	// show help info
	if (argc <= 1 || argc > 1 && (argv[1][0] == '?' ||
		(argv[1][0] == '-' || argv[1][0] == '/') &&
		(argv[1][1] == 'H' || argv[1][1] == 'h' || argv[1][1] == '?')))
	{
		showHelp(mypgm);
		return RETCODE_OK;
	}

#ifndef NOMPI
	try
	{
		MPI_Init(&argc, &argv);
		MPI_Comm_size(MPI_COMM_WORLD, &mpisize);
		MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
		int flag;
		MPI_Initialized(&flag);
		if (!flag)
			throw 1;
		else if (mpisize < 2)
		{
			MPI_Finalize();
			throw 2;
		}
	}
	catch (...)
	{
		mpisize = 1;
		myrank = 0;
	}
#endif  // ! NOMPI
    if (mpisize > 10)  rankw = 2;
    if (mpisize > 100) rankw = 3;

	t0 = getTimer();

	// copy arguments to the log
	if (myrank == 0)
	{
		plog << "\n" << mypgm << " v" << version;
		if (mpisize > 1)
			plog << " *" << mpisize;
		plog << ":";
		for (int i = 1; i < argc; i++)
			plog << " " << argv[i];
		cerr << plog.str() << endl;
		plog << endl;
	}

	string species_list;	// name of the species list

	// parse arguments
	for (int i = 1; i < argc; i++)
	{
		if (argv[i][0] == '-')				// parse options
		{
			char option = argv[i][1];
			char opt2 = argv[i][2];			// can be 0
			string::size_type s;
			string a;
			switch (option)
			{
				case '?':
				case 'H':
				case 'h':
					showHelp(mypgm);
				#ifndef NOMPI
					if (mpisize > 1) MPI_Finalize();
				#endif	// ifndef NOMPI
					return RETCODE_OK;
					break;
				case 'A':
				case 'a':
					if (opt2 == 0 && ++i < argc)
						aggregate = argv[i][0] == '0' ? 0 : 1;
					else if (opt2 == '0')
						aggregate = 0;
					else if (opt2 == '1')
						aggregate = 1;
					else
						argError(option);
					break;
				case 'B':
				case 'b':
					if (opt2 == 0 && ++i < argc)
						symmetry = argv[i][0] == '0' ? 0 : 1;
					else if (opt2 == '0')
						symmetry = 0;
					else if (opt2 == '1')
						symmetry = 1;
					else
						argError(option);
					break;
				case 'C':
				case 'c':
					if (opt2 == 0 && ++i < argc)
						usedup = argv[i][0] == '0' ? 0 : 1;
					else if (opt2 == '0')
						usedup = 0;
					else if (opt2 == '1')
						usedup = 1;
					else
						argError(option);
					break;
				case 'D':
				case 'd':
					if (++i < argc)
					{
						switch (opt2)
						{
						case 0:
							dupath.assign(argv[i]);
							s = dupath.find_first_of(star);
							if (s != string::npos)
								dupath.erase(s);
							else argError(option);

							dupext.assign(argv[i]);
							s = dupext.find_last_of(star);
							dupext.erase(0, s + 1);
							break;
						case 'D':
						case 'd':
						case 'P':
						case 'p':
							dupath.assign(argv[i]);
							break;
						case 'E':
						case 'e':
							dupext.assign(argv[i]);
							break;
						default: argError(option);
						}
					}
					else argError(option);
					break;
                case 'F':
                case 'f':
                    if (++i < argc)
                        ratio = (float)atof(argv[i]);
					else
						argError(option);
                    break;
				case 'G':
				case 'g':
					if (++i < argc)
					{
						switch (opt2)
						{
						case 0:
							genepath.assign(argv[i]);
							s = genepath.find_first_of(star);
							if (s != string::npos)
								genepath.erase(s);
							else argError(option);

							geneext.assign(argv[i]);
							s = geneext.find_last_of(star);
							geneext.erase(0, s + 1);
							break;
						case 'D':
						case 'd':
						case 'P':
						case 'p':
							genepath.assign(argv[i]);
							break;
						case 'E':
						case 'e':
							geneext.assign(argv[i]);
							break;
						default: argError(option);
						}
					}
					else argError(option);
					break;
				case 'I':
				case 'i':
					if (++i < argc)
					{
						switch (opt2)
						{
						case 0:
							inpath.assign(argv[i]);
							s = inpath.find_first_of(star);
							if (s != string::npos)
								inpath.erase(s);
							else argError(option);

							inext.assign(argv[i]);
							s = inext.find_last_of(star);
							inext.erase(0, s + 1);
							break;
						case 'D':
						case 'd':
						case 'P':
						case 'p':
							inpath.assign(argv[i]);
							break;
						case 'E':
						case 'e':
							inext.assign(argv[i]);
							break;
						default: argError(option);
						}
					}
					else argError(option);
					break;
				case 'K':
				case 'k':
					if (++i < argc)
					{
						switch (opt2)
						{
						case 'G':
						case 'g':
							gfield.assign(argv[i]);
							break;
						case 'P':
						case 'p':
							pfield.assign(argv[i]);
							break;
						case 'R':
						case 'r':
							rfield.assign(argv[i]);
							break;
						case 'S':
						case 's':
							sfield.assign(argv[i]);
							break;
						default: argError(option);
						}
					}
					else argError(option);
					break;
				case 'L':
				case 'l':
					if (++i < argc)
						logfile.assign(argv[i]);
					else
						argError(option);
					break;
				case 'M':
				case 'm':
					if (opt2 == 0 && ++i < argc)
						inmid.assign(argv[i]);
					else if (opt2 == 'M' || opt2 == 'm')
						inmid.clear();
					else argError(option);
					break;
				case 'N':
				case 'n':
					if (++i < argc)
					{
						a.assign(argv[i]);
						istringstream iss(a);
						iss >> maxhomo;
					}
					else argError(option);
					break;
				case 'O':
				case 'o':
					if (++i < argc)
					{
						switch (opt2)
						{
						case 0:
							outpath.assign(argv[i]);
							s = outpath.find_first_of(star);
							if (s != string::npos)
								outpath.erase(s);
							else argError(option);

							outext.assign(argv[i]);
							s = outext.find_last_of(star);
							outext.erase(0, s + 1);
							break;
						case 'D':
						case 'd':
						case 'P':
						case 'p':
							outpath.assign(argv[i]);
							break;
						case 'E':
						case 'e':
							outext.assign(argv[i]);
							break;
						default: argError(option);
						}
					}
					else argError(option);
					break;
				case 'R':
				case 'r':
                    if (opt2 == 0 && ++i < argc) {
                        addrank = (int)(argv[i][0] - '0');
                        if (addrank < 0 || addrank > 2)
                            argError(option);
                    }
					else if (opt2 == '0')
						addrank = 0;
                    else if (opt2 == '1')
                        addrank = 1;
                    else if (opt2 == '2')
                        addrank = 2;
					else
						argError(option);
					break;
				case 'S':
				case 's':
					if (++i < argc)
						hsortmode.assign(argv[i]);
					else argError(option);
					break;
				case 'T':
				case 't':
					if (++i < argc)
						vsortmode.assign(argv[i]);
					else argError(option);
					break;
				case 'V':
				case 'v':
					if (opt2 == 0 && ++i < argc)
						verbose = argv[i][0] == '0' ? 0 : 1;
					else if (opt2 == '0')
						verbose = 0;
					else if (opt2 == '1')
						verbose = 1;
					else
						argError(option);
					break;
				case 'W':
				case 'w':
					if (opt2 == 0 && ++i < argc)
						useweight = argv[i][0] == '0' ? 0 : 1;
					else if (opt2 == '0')
						useweight = 0;
					else if (opt2 == '1')
						useweight = 1;
					else
						argError(option);
					break;
				default:
					argError(option);
			}
		}       // blaster option parsed

		else if (species_list.empty())
			species_list.assign(argv[i]);

		else
		{
			string msg = "Extra positional argument ";
			msg.append(argv[i]);
			argError(msg);
		}
	}       // next argv

	// assign default values if necessary and refine args
	refineArguments();

	// prepare log file if required
	if (!logfile.empty())
	{
		det.open(logfile.c_str(), ios::app | ios::binary);
		if (det.bad())
		{
            if (mpisize > 1) cerr << "[" << right << setw(rankw) << myrank << "] ";
            cerr << "Can't open log file: " << logfile << endl;
		#ifndef NOMPI
			if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_EARG);
		#endif	// ! NOMPI
			return RETCODE_EARG;
		}
		det << plog.str() << flush;
	}

	// read species list
	if (species_list.empty())
	{
		if (myrank == 0)
		{
			string msg = "Posthit needs a list of species";
			cerr << msg << endl;
			if (det.is_open()) det << msg << endl;
		}
	#ifndef NOMPI
		if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_EARG);
	#endif	// ! NOMPI
		return RETCODE_EARG;
	}
	if (!readSpeciesList(species_list, species))
	{
		if (myrank == 0)
		{
			string msg = "Error in species list";
			cerr << msg << endl;
			if (det.is_open()) det << msg << endl;
		}
	#ifndef NOMPI
		if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_ELIST);
	#endif	// ! NOMPI
		return RETCODE_ELIST;
	}
	nsp = (int)species.size();

	if (myrank == 0)
	{
		ostringstream oss;
		oss << "\nTotal number of species: " << nsp << "\n\nReading protein&gene tables:";
		cerr << oss.str() << endl;
		if (det.is_open()) det << oss.str() << endl;
	}

	// read protein/gene tables for all species
	for (int sp = 0; sp < nsp; sp++)
	{
		if (!readGeneTable(sp))
		{
			if (myrank == 0)
			{ 
				string msg = "gene table read failed";
				cerr << msg << endl;
				if (det.is_open()) det << msg << endl;
			}
		#ifndef NOMPI
			if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_EINPUT);
		#endif	// ! NOMPI
			return RETCODE_EINPUT;
		}
	}

	if (myrank == 0)
	{
		ostringstream oss;
		oss << "\nTotal proteins=" << nprots << " unique=" << nunique << " group=" << nduplicated 
			<< " genes=" << ngenes;
		if (nambig > 0) oss << " ambig=" << nambig;
		if (nunknown > 0) oss << " unknown=" << nunknown; 
		cerr << oss.str() << endl;
		if (det.is_open()) det << oss.str() << endl;
	}

    // understand organization of hits
    if ( !understandHits())
    {   
        if (myrank==0)
		{
			string msg = "\nInvalid organization of hits";
			cerr << msg << endl;
			if (det.is_open()) det << msg << endl;
		}
	#ifndef NOMPI
		if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_EINPUT);
	#endif	// ! NOMPI
		return RETCODE_EINPUT;
	}

	// optionally fill self-scores for all proteins
	if (useweight)
	{
		if (myrank == 0)
		{
			string msg = "\nGathering self-scores for all proteins:";
			cerr << msg << endl;
			if (det.is_open()) det << msg << endl;
		}

		for (int sp = 0; sp < nsp; sp++)
		{
			if (!fillSelfScore(sp))
			{
				if (myrank == 0)
				{ 
					string msg = "filling self-score failed";
					cerr << msg << endl;
					if (det.is_open()) det << msg << endl;
				}
			#ifndef NOMPI
				if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_EINPUT);
			#endif	// ! NOMPI
				return RETCODE_EINPUT;
			}
		}

		// Count self-scores
		if (myrank == 0)
		{
			int cs = countSelf();
			ostringstream oss;
			oss << "\nSelf-scores found for " << cs << " of " << nunique + nduplicated << " proteins";
			cerr << oss.str() << endl;
			if (det.is_open()) det << oss.str() << endl;
		}
	}

	if (myrank == 0)
	{
		string msg = "\nFill score matrix of proteins/genes:";
		cerr << msg << endl;
		if (myrank == 0 && det.is_open()) det << msg << endl;
	}

	for (int sp = myrank; sp < nsp; sp+=mpisize)		// loop over 1st species in parallel
	{
		// fill prot->prot matrix of non-duplicates (with optional weights & symmetry)
		if (is_name2)   fillProtMatrix(sp);
        else            fillProtMatrix2(sp);

		// optionally expand the matrix with protein copies
		if (usedup) expandProtMatrix(sp);
		
		// optionally aggregate into the gene->gene matrix
		if (aggregate) aggregateGeneMatrix(sp);

		// sort and output the resulting files
		outputResult(sp);

		pmatrix.clear();
		gmatrix.clear();
	}

	// reduce results of all processes
#ifndef NOMPI
	if (mpisize > 1)
	{
		MPI_Barrier(MPI_COMM_WORLD);

		int exbuf1[3] = { total_rows, expand_rows, gene_rows }, inbuf1[3];
		MPI_Reduce(exbuf1, inbuf1, 3, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);
		
		long long exbuf2[3] = { total_cells, expand_cells, gene_cells }, inbuf2[3];
		MPI_Reduce(exbuf2, inbuf2, 3, MPI_LONG_LONG, MPI_SUM, 0, MPI_COMM_WORLD);

		if (myrank == 0)
		{
			total_rows  = inbuf1[0]; expand_rows  = inbuf1[1]; gene_rows  = inbuf1[2];
			total_cells = inbuf2[0]; expand_cells = inbuf2[1]; gene_cells = inbuf2[2];
		}
	}
#endif

    if (myrank == 0)
    {
        expand_rows += total_rows;
        expand_cells += total_cells;
        ostringstream oss;
        oss << "\nBuilt matrix of unique proteins: rows=" << total_rows << " cells=" << total_cells;
        if (usedup)
            oss << "\nExpanded with protein copies: rows=" << expand_rows << " cells=" << expand_cells;
        string what = aggregate == 0 ? "proteins" : "genes";
        oss << "\nThe matrix of " << what << " written: rows=" << gene_rows << " cells=" << gene_cells;
		cerr << oss.str() << endl;
		if (det.is_open()) det << oss.str() << endl;
    }

    if (addrank > 1)        // append backward ranks to all matrices
    {
        if (myrank == 0)
        {
            ostringstream oss;
            oss << "\nCompose common matrix of ranks:";
            cerr << oss.str() << endl;
            if (det.is_open()) det << oss.str() << endl;
        }

        // read matrices for all unexplored species in each process
        for (int sp = 0; sp < nsp; sp++)
        {
            if (sp % mpisize == myrank) continue;
            if (!readResult(sp))
            {
                if (myrank == 0)
                {
                    string msg = "reading result failed for species ";
                    cerr << msg << sp << endl;
                    if (det.is_open()) det << msg << sp << endl;
                }
            }
        }

#ifndef NOMPI
        if (mpisize > 1)
            MPI_Barrier(MPI_COMM_WORLD);
#endif

        if (myrank == 0)
        {
            ostringstream oss;
            oss << "\nAppend backward rank to the output matrices:";
            cerr << oss.str() << endl;
            if (det.is_open()) det << oss.str() << endl;
        }

        // loop over 1st species pertinent to this process
        for (int sp = myrank; sp < nsp; sp += mpisize)
        {
            if (!writeBackRank(sp))
            {
                if (myrank == 0)
                {
                    string msg = "appending backward rank failed for species ";
                    cerr << msg << sp << endl;
                    if (det.is_open()) det << msg << sp << endl;
                }
            }
        }
    }

#ifndef NOMPI
    if (mpisize > 1)
        MPI_Barrier(MPI_COMM_WORLD);
#endif

    if (myrank == 0)
    {
        ostringstream oss;
		t1 = getTimer() - t0;
		oss << "\nTime=" << fixed << setprecision(1) << t1 / 60.0 << "m";
		cerr << oss.str() << endl;
		if (det.is_open()) det << oss.str() << endl;
	}

#ifndef NOMPI
	if (mpisize > 1)
		MPI_Finalize();
#endif
}

void    showHelp (const string& program)
{		
	// free option letters: e f j p q u x y z
	cout << "\nUsage (v" << version << "): " << program << " [options] species_list\n"
		<< "\n"
		<< "Options (case insensitive):\n"
		<< "-a {0|1}   : 1=aggregate proteins to genes [" << def_aggregate << "], also -a0|-a1\n"
		<< "-b {0|1}   : 1=build symmetric matrices [" << def_symmetry << "], also -b0|-b1\n"
		<< "-c {0|1}   : 1=expand input with protein copies [" << def_usedup << "], also -c0|-c1\n"
		<< "-d string  : duplicates path & ext [" << def_dupath << "\\*" << def_dupext << "], also -dp,-de\n"
        << "-f float   : use (1-score/max) as rank below specified cutoff, 0=usual rank [" << def_ratio << "]\n"
		<< "-g string  : gene table path & ext [" << def_genepath << "\\*" << def_geneext << "], also -gp,-ge\n"
		<< "-h -?      : show this help information\n"
		<< "-i string  : input (hit) files path & ext [" << def_inpath << "\\*" << def_inext << "], also -ip,-ie\n"
		<< "-kg string : name of gene column in gene tables [" << def_gfield << "]\n"
		<< "-kp string : name of protein column in gene tables [" << def_pfield << "]\n"
		<< "-kr string : name of region column in gene tables [" << def_rfield << "]\n"
		<< "-ks string : name of start column in gene tables [" << def_sfield << "]\n"
		<< "-l string  : log file [none]\n"
		<< "-m string  : connector string between species names [" << def_inmid << "], -mm for none\n"
		<< "-n number  : max number of best homologs, 0=unlimited [" << def_maxhomo << "]\n"
		<< "-o string  : output files path & ext [" << def_outpath << "\\*" << def_outext << "], also -op,-oe\n"
		<< "-r {0|1|2} : include score rank(s) in output files [" << def_addrank << "], also -r0|-r1|-r2\n"
		<< "-s string  : horizontal sort mode: n i p s w e [" << def_hsortmode << "]\n"
		<< "-t string  : vertical sort mode: n i p s [" << def_vsortmode << "]\n"
		<< "-v {0|1}   : verbose output to console [" << def_verbose << "], also -v0|-v1\n"
		<< "-w {0|1}   : 1=convert score into weight [" << def_useweight << "], also -w0|-w1\n"
		<< "\n"
		<< "The line of the species_list should contain: species_name\\tFASTA_file_name\\t...\n"
		<< "Sort modes: n=none(unsorted) i=id p=position s=species w=weight/score e=E-value;\n"
		<< "            lower case letters = ascending order, upper case = descending order."
		<< "\n"
		<< "Examples: " << program << " -op result -oe .txt species.txt\n"
		<< "          " << program << " -l log.txt -w 1 species.txt\n"
		<< endl;
}

void	argError (char arg)
{
	if (myrank == 0)
	{
		string msg = "Error or missing value in option -";
		msg += arg;
		cerr << msg << endl;
		plog << msg << endl;
	}
#ifndef NOMPI
	if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_EARG);
#endif	// ! NOMPI                
	exit(RETCODE_EARG);
}

void	argError (const string & msg)
{
	if (myrank == 0)
	{
		cerr << msg << endl;
		plog << msg << endl;
	}
#ifndef NOMPI
	if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_EARG);
#endif	// ! NOMPI                
	exit(RETCODE_EARG);
}

// assign default values if necessary and refine args
void	refineArguments (void)
{
	if (inpath == unspec)
		inpath.assign(def_inpath);
	if (!inpath.empty() && *inpath.rbegin() != slash && *inpath.rbegin() != slash2)
		inpath.append(1, slash);

	if (inmid == unspec)
		inmid.assign(def_inmid);

	if (inext == unspec)
		inext.assign(def_inext);
	if (!inext.empty() && inext[0] != '.')
		inext = '.' + inext;

	if (dupath == unspec)
		dupath.assign(def_dupath);
	if (!dupath.empty() && *dupath.rbegin() != slash && *dupath.rbegin() != slash2)
		dupath.append(1, slash);

	if (dupext == unspec)
		dupext.assign(def_dupext);
	if (!dupext.empty() && dupext[0] != '.')
		dupext = '.' + dupext;

	if (genepath == unspec)
		genepath.assign(def_genepath);
	if (!genepath.empty() && *genepath.rbegin() != slash && *genepath.rbegin() != slash2)
		genepath.append(1, slash);

	if (geneext == unspec)
		geneext.assign(def_geneext);
	if (!geneext.empty() && geneext[0] != '.')
		geneext = '.' + geneext;

	if (outpath == unspec)
		outpath.assign(def_outpath);
	if (!outpath.empty() && *outpath.rbegin() != slash && *outpath.rbegin() != slash2)
		outpath.append(1, slash);
	ensureDirectory(outpath);

	if (outext == unspec)
		outext.assign(def_outext);
	if (!outext.empty() && outext[0] != '.')
		outext = '.' + outext;

	if (!vsortmode.empty())
	{
		if (vsortmode.find_first_of("nN") != string::npos)
			vsortmode.clear();
		else	// erase unknown characters
			for (int i = (int)vsortmode.size() - 1; i >= 0; i--)
			{
				if (vmodes.find(vsortmode[i]) == string::npos)
					vsortmode.erase(i, 1);
			}
	}

	if (!hsortmode.empty())
	{
		if (hsortmode.find_first_of("nN") != string::npos)
			hsortmode.clear();
		else	// erase unknown characters
			for (int i = (int)hsortmode.size() - 1; i >= 0; i--)
			{
				if (hmodes.find(hsortmode[i]) == string::npos)
					hsortmode.erase(i, 1);
			}
	}
    if (ratio != 0 && hsortmode != def_hsortmode)
    {
        if (myrank == 0)
        {
            string msg = "Cannot use ratio for rank: nonstandard horisontal ordering.";
            cerr << msg << endl;
            plog << msg << endl;
        }
    }
}

bool	ensureDirectory(const string& dir)
{
	if (isFile(dir.c_str()) == 0)
		return true;
#ifdef WINDOWS
	_flushall();
	string cmd = "mkdir " + dir + " 2>nul";
#else
	string cmd = "mkdir -p -m 0775 " + dir;
#endif	// WINDOWS
	int rc = system(cmd.c_str());
	int err = (rc == 0) ? 0 : errno;
	return (err == 0 || err == EEXIST || err == ENOENT);
}

bool    readSpeciesList(const string& file, vector <Species>& species)
{
	string::size_type pos;
	ifstream    lst;

	lst.open(file.c_str(), ios_base::binary);
	if (!lst.good())
	{
		if (myrank == 0)
		{
			string str = "Cannot open list of species: " + file;
			cerr << str << endl;
			if (det.is_open()) det << str << endl;
		}
		return false;
	}

	int lineno = 1;
	for (; !lst.eof(); lineno++)
	{
		lst.getline(buf, buflen);
		if (lst.fail())
		{
			if (lst.eof()) break;
			if (myrank == 0)
			{
				ostringstream oss;
				oss << "Error ignored in " << file << " at line " << lineno;
				cerr << oss.str() << endl;
				if (det.is_open())
					det << oss.str() << endl;
			}
			continue;
		}

        //skip header if present
        if (lineno == 1 && buf[0] == ' ') continue;

		Species spec;
		string line(buf);

		if (line.empty())
			continue;
		if (comment.find(line[0]) != string::npos)
			continue;
        if (*line.rbegin() == '\r')
            line.erase(line.size() - 1);

		// parse alias (species name)
		pos = line.find_first_of(tab);
		//if (pos == string::npos)
		//{
		//	if (myrank == 0)
		//	{
		//		ostringstream oss;
		//		oss << " no 2nd field in line " << lineno;
		//		cerr << oss.str() << endl;
		//		if (det.is_open())
		//			det << oss.str() << endl;
		//	}
		//	return false;
		//}
		//else
			spec.name = line.substr(0, pos);

		// parse filename - not needed
		//off = line.find_first_not_of(tab, pos);
		//if (off == string::npos)
		//{
		//	if (myrank == 0)
		//	{
		//		ostringstream oss;
		//		oss << "no file name in line " << lineno;
		//		cerr << oss.str() << endl;
		//		if (det.is_open())
		//			det << oss.str() << endl;
		//	}
		//	return false;
		//}
		//else
		//{
		//	pos = line.find_first_of(tab, off);
		//	if (pos == string::npos)
		//		spec.fasta = line.substr(off);
		//	else
		//		spec.fasta = line.substr(off, pos - off);
		//}

		species.push_back(spec);
	}

	lst.close();
	return true;
}

bool	readGeneTable (const int sp)
{
	string filename = genepath + species[sp].name + geneext;
	if (myrank == 0)
	{
		ostringstream oss;
		oss << left << setw(maxw) << species[sp].name + "... " << right;
		cerr << oss.str() << flush;
		if (det.is_open()) det << oss.str() << flush;
	}

	ifstream gtab;
	gtab.open(filename.c_str(), ios_base::binary);
	if (!gtab.good())
	{
		if (myrank == 0)
		{
			string msg = filename + " open failed";
			cerr << msg << endl;
			if (det.is_open()) det << msg << endl;
		}
		return false;
	}

	gtab.getline(buf, buflen);		// read heading
	string line(buf);

	// finding protein and gene columns
	int pfno = findField(line, pfield);
	int gfno = findField(line, gfield);
	int rfno = findField(line, rfield);
	int sfno = findField(line, sfield);
	if (pfno < 0 || gfno < 0 || rfno < 0 || sfno < 0)
	{
		if (myrank == 0)
		{
			string msg = "field(s) not found";
			cerr << msg << endl;
			if (det.is_open()) det << msg << endl;
		}
		return false;
	}

	// read protein & gene IDs and coordinates
	int np = 0, ng = 0, na = 0;		// local count of proteins, genes, and ambiguities thereof
	int lineno = 1;
	while (!gtab.eof())
	{
		gtab.getline(buf, buflen);
		if (!gtab.good()) break;
		line.assign(buf);
		if (line.empty()) break;
		lineno++;

		Protein prot;
		prot.sp = sp;
		Gene gene;
		gene.sp = sp;
		string::size_type pos, off = 0;

		for (int i = 0; i <= pfno || i <= gfno || i <= rfno || i <= sfno; i++)
		{
			pos = line.find_first_of(tab, off);
			string f;
			if (pos == string::npos)	f = line.substr(off);
			else						f = line.substr(off, pos - off);

			if (i == pfno)		prot.id = f;
			else if (i == gfno)	gene.id = f;
			else if (i == rfno) gene.region = f;
			else if (i == sfno)
			{
				istringstream iss(f);
				iss >> gene.start;
			}

			off = pos + 1;
			if (off >= line.size()) break;
		}

		if (prot.id.empty() || gene.id.empty()) 
			continue;			// skip if a protein id or gene id miss

		// try insert new protein
		pair <DictIt, bool> res;
		res = prodict.insert(DictVal(prot.id, nprots));

		if (!res.second)		// protein repeat, skip line
		{
			if (myrank == 0 && verbose != 0)
			{
				string cmpgene = "same";
				int p = res.first->second;
				int g = proteins[p].gno;
				if (genes[g].id != gene.id)
					cmpgene = "different " + genes[g].id;
				ostringstream oss;
				oss << "protein " << prot.id << " repeated with " << cmpgene << " gene in line " << lineno;
				cerr << oss.str() << endl;
				if (det.is_open()) det << oss.str() << endl;
			}
			na++;
			nambig++;
			continue;
		}

		// try insert new gene  *NB: In theory, a protein may refer to multiple genes
		res = gendict.insert(DictVal(gene.id, ngenes));

		if (!res.second)		// gene repeat, store its number
		{
			int gno = res.first->second;
			prot.gno = gno;
			genes[gno].pno.push_back(nprots);
		}
		else					// add new gene
		{
			prot.gno = ngenes;
			gene.pno.push_back(nprots);
			genes.push_back(gene);
			ngenes++;
			ng++;		// of this species
		}

		proteins.push_back(prot);
		nprots++;
		np++;		// of this species
	}

	gtab.close();
	int nlines = lineno;

	// read the table of duplicates
	filename = dupath + species[sp].name + dupext;
	ifstream fdup;
	fdup.open(filename.c_str(), ios_base::binary);
	if (!fdup.good())
	{
		if (myrank == 0)
		{
			string msg = filename + " open failed";
			cerr << msg << endl;
			if (det.is_open()) det << msg << endl;
		}
		return false;
	}

	int nn = 0, nu = 0, nd = 0;		// local count of unknown, unique and duplicated proteins (excl. copies)
	lineno = 0;
	while (!fdup.eof())
	{
		fdup.getline(buf, buflen);
		if (!fdup.good()) break;
		line.assign(buf);
		if (line.empty()) break;
		lineno++;

		string protid;
		int ncopies;
		istringstream iss(line);
		iss >> protid >> ncopies;

		DictIt mit = prodict.find(protid);
		if (mit == prodict.end())			// master id not found
		{
			if (myrank == 0 && verbose != 0)
			{
				ostringstream oss;
				oss << "master protein " << protid << " not found in line " << lineno;
				cerr << oss.str() << endl;
				if (det.is_open()) det << oss.str() << endl;
			}
			nn++;
			nunknown++;
			continue;
		}

		nd++;
		nduplicated++;
		int master_no = mit->second;
		Protein& master = proteins[master_no];

		for (int k = 0; k < ncopies; k++)
		{
			iss >> protid;
			mit = prodict.find(protid);
			if (mit == prodict.end())			// duplicate id not found
			{
				if (myrank == 0 && verbose != 0)
				{
					ostringstream oss;
					oss << "duplicate protein " << protid << " not found in line " << lineno;
					cerr << oss.str() << endl;
					if (det.is_open()) det << oss.str() << endl;
				}
				nn++;
				nunknown++;
				continue;
			}
			master.dup.push_back(mit->second);
			proteins[mit->second].master = master_no;
		}
	}
	fdup.close();

	for (int i = 0; i < (int)proteins.size(); i++)
	{
		Protein& p = proteins[i];
		if (p.sp == sp && p.master < 0 && p.dup.empty())
		{
			nu++;
			nunique++;
		}
	}

	if (myrank == 0)
	{
		ostringstream oss;
		oss << "lin=" << left << setw(6) << nlines << " pro=" << setw(6) << np << " uni=" 
			<< setw(6) << nu << " grp=" << setw(5) << nd << " gen=" << setw(6) << ng;
		if (na != 0) oss << " amb=" << na;
		if (nn != 0) oss << " unk=" << nn;
		cerr << oss.str() << endl;
		if (det.is_open()) det << oss.str() << endl;
	}

	return true;
}

bool    understandHits (void)  
{   
    if (nsp < 2)    return false;
    Species& spec1 = species[0];
    Species& spec2 = species[1];

    string f1 = inpath + spec1.name + slash;                    // hits/spec/ ?
    string f2 = inpath + spec2.name + slash;
    if (isFile(f1.c_str()) != 0 || isFile(f2.c_str()) != 0)     // no
    {
        is_subdir = false;
        f1 = inpath + spec1.name + inmid + spec2.name + inext;              // hits/spec1-spec2.ext ?
        f2 = inpath + spec2.name + inmid + spec1.name + inext;
        if (isFile(f1.c_str()) != 0 || isFile(f2.c_str()) != 0)     // no
        {
            f1 = inpath + spec1.name + inext;                           // hits/spec.ext ?
            f2 = inpath + spec2.name + inext;
            if (isFile(f1.c_str()) == 0 && isFile(f2.c_str()) == 0)     // yes
            {
                is_name2 = false;
                return true;
            }
            else return false;                                          // no
        }
        else return true;                                           // yes
    }

    else                                                        // yes
    {
        f1 += spec1.name + inmid + spec2.name + inext;              // hits/spec1/spec1-spec2.ext ?
        f2 += spec2.name + inmid + spec1.name + inext;
        if (isFile(f1.c_str()) != 0 || isFile(f2.c_str()) != 0)     // no
        {
            f1 = inpath + spec1.name + slash + spec2.name + inext;      // hits/spec1/spec2.ext ?
            f2 = inpath + spec2.name + slash + spec1.name + inext;
            if (isFile(f1.c_str()) == 0 && isFile(f2.c_str()) == 0)     // yes
            {
                is_name1 = false;
                return true;
            }
            else return false;                                          // no
        }
        else return true;                                           // yes
    }

    return true;
}

// return zero based number of given field within a line or RETCODE_NONE
int		findField (const string& str, const string& field)
{
	string::size_type pos, off = 0;

	for (int i = 0; ; i++)
	{
		pos = str.find_first_of(tab, off);
		if (pos == string::npos)
			return (str.substr(off) == field) ? i : RETCODE_NONE;
		else if (str.substr(off, pos - off) == field)
			return i;
		off = pos + 1;
		if (off >= str.size()) break;
	}
	return RETCODE_NONE;
}

// fetch self-scores for all proteins
bool	fillSelfScore(const int sp)
{
	Species& spec = species[sp];
    string filename = inpath;
    if (is_subdir)  filename += spec.name + slash;
    if (is_name1)   filename += spec.name;
    if (is_name1 && is_name2)   filename += inmid;
    if (is_name2)   filename += spec.name;
    filename += inext;

    if (myrank == 0)
	{
		ostringstream oss;
		oss << left << setw(maxw) << spec.name + "... " << right;
		cerr << oss.str() << flush;
		if (det.is_open()) det << oss.str() << flush;
	}

	ifstream sco;
	sco.open(filename.c_str(), ios_base::binary);
	if (!sco.good())
	{
		if (myrank == 0)
		{
			string msg = filename + " open failed";
			cerr << msg << endl;
			if (det.is_open()) det << msg << endl;
		}
		return false;
	}

	// hit reading loop
	int np = 0;				// number of proteins processed
	int nerr = 0;			// number of errors encountered
	int lineno = 0;
	float smin = 1.e10;
	float smax = 0.0;
	double ssum = 0.0;
	string line;
	while (!sco.eof())
	{
		sco.getline(buf, buflen);
		if (!sco.good()) break;
		line.assign(buf);
		if (line.empty()) break;
		lineno++;

		string id, ww;
		float score;
		string::size_type pos, off = 0;

		pos = line.find_first_of(tab, off);		// ID1
		if (pos == string::npos)	id = line.substr(off);
		else						id = line.substr(off, pos - off);
		if (id.empty()) continue;
		off = pos + 1;

		pos = line.find_first_of(tab, off);		// ID2
		if (pos == string::npos)	ww = line.substr(off);
		else						ww = line.substr(off, pos - off);
		if (id != ww) continue;
		off = pos + 1;

		pos = line.find_first_of(tab, off);		// Score
		if (pos == string::npos)	ww = line.substr(off);
		else						ww = line.substr(off, pos - off);
		if (ww.empty()) continue;

		istringstream iss(ww);
		iss >> score;

		// find protein and change its self-score
		DictIt dit = prodict.find(id);
		if (dit == prodict.end())
		{
			if (myrank == 0 && verbose != 0)
			{
				ostringstream oss;
				oss << "unknown protein " << id << " found in line " << lineno;
				cerr << oss.str() << endl;
				if (det.is_open())
					det << oss.str() << endl;
			}
			nerr++;
		}
		else
		{
			if (smin > score) smin = score;
			if (smax < score) smax = score;

			Protein & prot = proteins[dit->second];
			if (prot.master > 0)  // this is a duplicate
			{
				if (proteins[prot.master].self > 0)
					continue;
				else
				{
					np++;
					ssum += score;
					proteins[prot.master].self = score;
				}
			}

			else if (prot.self == 0.0)
			{
				np++;
				ssum += score;
				prot.self = score;
			}
			else if (score > prot.self)
			{
				ssum -= prot.self;
				ssum += score;
				prot.self = score;
			}
		}
	}

	if (myrank == 0)
	{
		ostringstream oss;
		oss << "lin=" << lineno << "  pro=" << np << "  sco=" << fixed << setprecision(0) 
			<< smin << ".." << smax << "(" << ssum/(np + 1e-10) << ")";
		if (nerr > 0) oss << "  err=" << nerr;
		cerr << oss.str() << endl;
		if (det.is_open())
			det << oss.str() << endl;
	}

	sco.close();
	return true;
}

// return number of nonduplicate proteins that have nonzero self-score
int		countSelf(void)
{
	int rc = 0;
	for (int p = 0; p < nprots; p++)
	{
		Protein& prot = proteins[p];
		if (prot.master < 0 && prot.self > 0.0) rc++;
	}
	return rc;
}

// fill non-duplicate protein matrix from hit files of given 1st species
void	fillProtMatrix (const int sp)
{
    assert(is_name2);
	ostringstream oss;
	int nrows = 0, ncells = 0, nerrors = 0;
	Species& spec = species[sp];

	for (int s2 = 0; s2 < nsp; s2++)
	{
		Species& spec2 = species[s2];
        string filename = inpath;
        if (is_subdir)  filename += spec.name + slash;
        if (is_name1)   filename += spec.name + inmid;
        filename += spec2.name + inext;

		if (verbose != 0)
		{
			ostringstream spa;
            if (mpisize > 1) spa << "[" << right << setw(rankw) << myrank << "] "; 
            spa << spec.name << " vs " << spec2.name;
			oss << left << setw(2 * maxw) << spa.str() << right << ":";
		}

		ifstream sco;
		sco.open(filename.c_str(), ios_base::binary);
		if (!sco.good())
		{
			if (verbose != 0)
			{
				oss << "open failed";
				cerr << oss.str() << endl;
				if (det.is_open()) det << oss.str() << endl;
			}
			continue;
		}

		// hit reading loop
		int nr = 0;				// number of rows added
		int nc = 0;				// number of cells added
		int nerr = 0;			// number of errors encountered
		int lineno = 0;
		string line;
		DictIt dit;

		while (!sco.eof())
		{
			sco.getline(buf, buflen);
			if (!sco.good()) break;
			line.assign(buf);
			if (line.empty()) break;
			lineno++;

			string id1, id2;
			double score, evalue;
			istringstream iss(line);
			iss >> id1 >> id2 >> score >> evalue;

			dit = prodict.find(id1);
			if (dit == prodict.end()) { nerr++; continue; }

			int pno1 = dit->second;
			if (proteins[pno1].master >= 0)
				pno1 = proteins[pno1].master;
			Protein & prot1 = proteins[pno1];

			dit = prodict.find(id2);
			if (dit == prodict.end()) { nerr++; continue; }

			int pno2 = dit->second;
			if (proteins[pno2].master >= 0)
				pno2 = proteins[pno2].master;
			Protein & prot2 = proteins[pno2];

			if (useweight)			// convert score into weight if needed
				score = (score + score) / (prot1.self + prot2.self + nonzero);

			if (prot1.rno >= 0)		// row already exists
			{
				Row & row = pmatrix[prot1.rno];
				if (row.ino != pno1)		// internal error in the matrix
				{
				#ifndef NOMPI
					if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_EMATRIX*10+1);
				#endif
					exit(RETCODE_EMATRIX*10+1);
				}

				// find cell(pno1, pno2) if exists in this row
				int c = 0;
				for (; c < (int)row.cell.size(); c++)
				{
					Cell & cell = row.cell[c];
					if (cell.jno == pno2)			// cell already exists
					{
						if (cell.eval > evalue) cell.eval = evalue;
						if (cell.score < (float)score) cell.score = (float)score;
						break;
					}
				}		// try next cell

				if (c >= (int)row.cell.size())		// add cell(pno1, pno2) to this row
				{
					Cell cell;
					cell.jno = pno2;
					cell.score = (float)score;
					cell.eval = evalue;
					row.cell.push_back(cell);
					nc++;
				}		// add cell
			}		// row exists

			else					// create row(pno1) and cell(pno1, pno2)
			{
				Row row;
				row.ino = pno1;

				Cell cell;
				cell.jno = pno2;
				cell.score = (float)score;
				cell.eval = evalue;

				row.cell.push_back(cell);
				prot1.rno = (int)pmatrix.size();
				pmatrix.push_back(row);
				nc++; nr++;
			}		// row and cell added
		}		// next hit

		sco.clear(); sco.close(); 

		// process inverse matrix block
		if (symmetry && s2 != sp) {			

            filename = inpath;
            if (is_subdir)  filename += spec2.name + slash;
            if (is_name1)   filename += spec2.name + inmid; 
            filename += spec.name + inext;

			sco.open(filename.c_str(), ios_base::binary);
			if (sco.good())		// file exists and good
			{
				lineno = 0;
				while (!sco.eof())
				{
					sco.getline(buf, buflen);
					if (!sco.good()) break;
					line.assign(buf);
					if (line.empty()) break;
					lineno++;

					string id1, id2;
					double score, evalue;
					istringstream iss(line);
					iss >> id2 >> id1 >> score >> evalue;		// inverse ID order!

					dit = prodict.find(id1);
					if (dit == prodict.end()) { nerr++; continue; }

					int pno1 = dit->second;
					if (proteins[pno1].master >= 0)
						pno1 = proteins[pno1].master;
					Protein & prot1 = proteins[pno1];

					dit = prodict.find(id2);
					if (dit == prodict.end()) { nerr++; continue; }

					int pno2 = dit->second;
					if (proteins[pno2].master >= 0)
						pno2 = proteins[pno2].master;
					Protein & prot2 = proteins[pno2];

					if (useweight)			// convert score into weight if needed
						score = (score + score) / (prot1.self + prot2.self + nonzero);

					if (prot1.rno >= 0)		// row already exists
					{
						Row & row = pmatrix[prot1.rno];
						if (row.ino != pno1)		// internal error in the matrix
						{
						#ifndef NOMPI
							if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_EMATRIX*10+2);
						#endif
							exit(RETCODE_EMATRIX*10+2);
						}

						// find cell(pno1, pno2) if exists in this row
						int c = 0;
						for (; c < (int)row.cell.size(); c++)
						{
							Cell & cell = row.cell[c];
							if (cell.jno == pno2)			// cell already exists
							{
								if (cell.eval > evalue) cell.eval = evalue;
								if (cell.score < (float)score) cell.score = (float)score;
								break;
							}
						}		// try next cell

						if (c >= (int)row.cell.size())		// add cell(pno1, pno2) to this row
						{
							Cell cell;
							cell.jno = pno2;
							cell.score = (float)score;
							cell.eval = evalue;
							row.cell.push_back(cell);
							nc++;
						}		// add cell
					}		// row exists

					else					// create row(pno1) and cell(pno1, pno2)
					{
						Row row;
						row.ino = pno1;

						Cell cell;
						cell.jno = pno2;
						cell.score = (float)score;
						cell.eval = evalue;

						row.cell.push_back(cell);
						prot1.rno = (int)pmatrix.size();
						pmatrix.push_back(row);
						nc++; nr++;
					}		// row and cell added
				}		// next inverse hit

				sco.clear(); sco.close();
			}
		}

		if (verbose != 0)
		{
			oss << " row=" << nr << " cell=" << nc << " err=" << nerr;
			cerr << oss.str() << endl;
			if (det.is_open()) det << oss.str() << endl;
		}
		oss.str(""); oss.clear();

		nrows += nr;
		ncells += nc;
		nerrors += nerr;
	}		// next 2nd species

	ostringstream osp;
    if (mpisize > 1) osp << "[" << right << setw(rankw) << myrank << "] "; 
    osp << spec.name;
	oss << left << setw(maxw) << osp.str() << right << " row=" << setw(5) 
		<< nrows << "  cell=" << setw(7) << ncells << "  err=" << nerrors;
	cerr << oss.str() << endl;
	if (det.is_open()) det << oss.str() << endl;
	total_rows += nrows;
	total_cells += ncells;
}

// fill non-duplicate protein matrix in the case of is_name2 = false
void    fillProtMatrix2(const int sp)
{
    assert(is_name1 && !is_name2);
    ostringstream oss;
    int nrows = 0, ncells = 0, nerrors = 0;
    Species& spec = species[sp];
    string filename = inpath;
    if (is_subdir)  filename += spec.name + slash;
    filename += spec.name + inext;

    if (verbose != 0)
    {
        ostringstream spa;
        if (mpisize > 1) spa << "[" << right << setw(rankw) << myrank << "] ";
        spa << spec.name << " vs all";
        oss << left << setw(maxw + 5) << spa.str() << right << ":";
    }

    ifstream sco;
    sco.open(filename.c_str(), ios_base::binary);
    if (!sco.good())
    {
        if (verbose != 0)
        {
            oss << "open failed";
            cerr << oss.str() << endl;
            if (det.is_open()) det << oss.str() << endl;
        }
        return;
    }

    // hit reading loop
    int nr = 0;				// number of rows added
    int nc = 0;				// number of cells added
    int nerr = 0;			// number of errors encountered
    int lineno = 0;
    string line;
    DictIt dit;

    while (!sco.eof())
    {
        sco.getline(buf, buflen);
        if (!sco.good()) break;
        line.assign(buf);
        if (line.empty()) break;
        lineno++;

        string id1, id2;
        double score, evalue;
        istringstream iss(line);
        iss >> id1 >> id2 >> score >> evalue;

        dit = prodict.find(id1);
        if (dit == prodict.end()) { nerr++; continue; }

        int pno1 = dit->second;
        if (proteins[pno1].master >= 0)
            pno1 = proteins[pno1].master;
        Protein & prot1 = proteins[pno1];

        dit = prodict.find(id2);
        if (dit == prodict.end()) { nerr++; continue; }

        int pno2 = dit->second;
        if (proteins[pno2].master >= 0)
            pno2 = proteins[pno2].master;
        Protein & prot2 = proteins[pno2];

        if (useweight)			// convert score into weight if needed
            score = (score + score) / (prot1.self + prot2.self + nonzero);

        if (prot1.rno >= 0)		// row already exists
        {
            Row & row = pmatrix[prot1.rno];
            if (row.ino != pno1)		// internal error in the matrix
            {
#ifndef NOMPI
                if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_EMATRIX * 10 + 1);
#endif
                exit(RETCODE_EMATRIX * 10 + 1);
            }

            // find cell(pno1, pno2) if exists in this row
            int c = 0;
            for (; c < (int)row.cell.size(); c++)
            {
                Cell & cell = row.cell[c];
                if (cell.jno == pno2)			// cell already exists
                {
                    if (cell.eval > evalue) cell.eval = evalue;
                    if (cell.score < (float)score) cell.score = (float)score;
                    break;
                }
            }		// try next cell

            if (c >= (int)row.cell.size())		// add cell(pno1, pno2) to this row
            {
                Cell cell;
                cell.jno = pno2;
                cell.score = (float)score;
                cell.eval = evalue;
                row.cell.push_back(cell);
                nc++;
            }		// add cell
        }		// row exists

        else					// create row(pno1) and cell(pno1, pno2)
        {
            Row row;
            row.ino = pno1;

            Cell cell;
            cell.jno = pno2;
            cell.score = (float)score;
            cell.eval = evalue;

            row.cell.push_back(cell);
            prot1.rno = (int)pmatrix.size();
            pmatrix.push_back(row);
            nc++; nr++;
        }		// row and cell added
    }		// next hit

    sco.clear(); sco.close();

    // process inverse matrices: only blocks relevant to sp
    if (symmetry)
    {
        for (int s2 = 0; s2 < nsp; s2++)
        {
            if (s2 == sp) continue;
            Species& spec2 = species[s2];
            filename = inpath;
            if (is_subdir)  filename += spec2.name + slash;
            filename += spec2.name + inext;

            sco.open(filename.c_str(), ios_base::binary);
            if (!sco.good())		// file exists and good
            {
                if (verbose != 0)
                {
                    oss << filename << " etc. open failed";
                    cerr << oss.str() << endl;
                    if (det.is_open()) det << oss.str() << endl;
                }
                break;
            }

            // backward hit reading loop
            while (!sco.eof())
            {
                sco.getline(buf, buflen);
                if (!sco.good()) break;
                line.assign(buf);
                if (line.empty()) break;
                lineno++;

                string id1, id2;
                double score, evalue;
                istringstream iss(line);
                iss >> id2 >> id1 >> score >> evalue;		// inverse ID order!

                dit = prodict.find(id1);
                if (dit == prodict.end()) { nerr++; continue; }

                int pno1 = dit->second;
                if (proteins[pno1].master >= 0)
                    pno1 = proteins[pno1].master;
                Protein & prot1 = proteins[pno1];

                if (prot1.sp != sp) continue;       // skip hits to other species' proteins

                dit = prodict.find(id2);
                if (dit == prodict.end()) { nerr++; continue; }

                int pno2 = dit->second;
                if (proteins[pno2].master >= 0)
                    pno2 = proteins[pno2].master;
                Protein & prot2 = proteins[pno2];

                if (useweight)			// convert score into weight if needed
                    score = (score + score) / (prot1.self + prot2.self + nonzero);

                if (prot1.rno >= 0)		// row already exists
                {
                    Row & row = pmatrix[prot1.rno];
                    if (row.ino != pno1)		// internal error in the matrix
                    {
#ifndef NOMPI
                        if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_EMATRIX * 10 + 2);
#endif
                        exit(RETCODE_EMATRIX * 10 + 2);
                    }

                    // find cell(pno1, pno2) if exists in this row
                    int c = 0;
                    for (; c < (int)row.cell.size(); c++)
                    {
                        Cell & cell = row.cell[c];
                        if (cell.jno == pno2)			// cell already exists
                        {
                            if (cell.eval > evalue) cell.eval = evalue;
                            if (cell.score < (float)score) cell.score = (float)score;
                            break;
                        }
                    }		// try next cell

                    if (c >= (int)row.cell.size())		// add cell(pno1, pno2) to this row
                    {
                        Cell cell;
                        cell.jno = pno2;
                        cell.score = (float)score;
                        cell.eval = evalue;
                        row.cell.push_back(cell);
                        nc++;
                    }		// add cell
                }		// row exists

                else					// create row(pno1) and cell(pno1, pno2)
                {
                    Row row;
                    row.ino = pno1;

                    Cell cell;
                    cell.jno = pno2;
                    cell.score = (float)score;
                    cell.eval = evalue;

                    row.cell.push_back(cell);
                    prot1.rno = (int)pmatrix.size();
                    pmatrix.push_back(row);
                    nc++; nr++;
                }		// row and cell added
            }		// next inverse hit

            sco.clear(); sco.close();
        }       // next inverse file
    }

    if (verbose != 0)
    {
        oss << " row=" << nr << " cell=" << nc << " err=" << nerr;
        cerr << oss.str() << endl;
        if (det.is_open()) det << oss.str() << endl;
    }
    oss.str(""); oss.clear();

    nrows += nr;
    ncells += nc;
    nerrors += nerr;

    ostringstream osp;
    if (mpisize > 1) osp << "[" << right << setw(rankw) << myrank << "] ";
    osp << spec.name;
    oss << left << setw(maxw) << osp.str() << right << " row=" << setw(5)
        << nrows << "  cell=" << setw(7) << ncells << "  err=" << nerrors;
    cerr << oss.str() << endl;
    if (det.is_open()) det << oss.str() << endl;
    total_rows += nrows;
    total_cells += ncells;
}

// expand the protein score matrix with duplicate proteins
void	expandProtMatrix (const int sp)
{
	ostringstream oss;
	int plus_row = 0;
	int plus_cell = 0;
	Species& spec = species[sp];

	// first expand cell vector of each row
	int rmax = (int)pmatrix.size();

	for (int i = 0; i < rmax; i++)
	{
		Row & row = pmatrix[i];
		if (proteins[row.ino].sp != sp)
			continue;

		int cmax = (int)row.cell.size();

		for (int j = 0; j < cmax; j++)
		{
			int pno2 = row.cell[j].jno;

			Protein & prot2 = proteins[pno2];
			if (prot2.master >= 0)
			{
			#ifndef NOMPI
				if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_EMATRIX*10+3);
			#endif
				exit(RETCODE_EMATRIX*10+3);
			}

			for (int k = 0; k < (int)prot2.dup.size(); k++)
			{
				Cell & cell = row.cell[j];
				Cell c;
				c.jno   = prot2.dup[k];
				c.score = cell.score;
				c.eval  = cell.eval;
				row.cell.push_back(c);
				plus_cell ++;
			}		// next copy of protein2
		}		// next cell of the current row

	}		// next row

	// then expand rows with protein copies
	for (int i = 0; i < rmax; i++)
	{
		int pno1 = pmatrix[i].ino;
		if (proteins[pno1].sp != sp)
			continue;

		Protein & prot1 = proteins[pno1];
		if (prot1.master >= 0)
		{
		#ifndef NOMPI
			if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_EMATRIX*10+4);
		#endif
			exit(RETCODE_EMATRIX*10+4);
		}

		if (prot1.dup.empty()) continue;

		for (int k = 0; k < (int)prot1.dup.size(); k++)
		{
			Row & row = pmatrix[i];
			Row r;
			r.ino = prot1.dup[k];
			r.cell.assign(row.cell.begin(), row.cell.end());

			pmatrix.push_back(r);
			plus_row ++;
			plus_cell += (int)r.cell.size();
		}		// next duplicate of protein1
	}		// next row, i.e. protein1

	expand_rows += plus_row;
	expand_cells += plus_cell;

	ostringstream osp;
    if (mpisize > 1) osp << "[" << right << setw(rankw) << myrank << "] "; 
    osp << spec.name;
	oss << left << setw(maxw) << osp.str() << right << "    +" << setw(5) 
		<< plus_row << "      +" << setw(7) << plus_cell;
	cerr << oss.str() << endl;
	if (det.is_open()) det << oss.str() << endl;
}

// aggregate pmatrix into gmatrix (gene matrix)
void	aggregateGeneMatrix (const int sp)
{
	Species& spec = species[sp];

	// select protein rows in turn
	for (int prowno = 0; prowno < (int)pmatrix.size(); prowno++)
	{
		Row & prow = pmatrix[prowno];
		Protein & prot1 = proteins[prow.ino];
		if (prot1.sp != sp) continue;

		int gno1 = prot1.gno;
		int growno = -1;
		IndexIt grit = gmindex.find(gno1);

		if (grit == gmindex.end())				// create new gene row
		{
			growno = (int)gmatrix.size();
			Row grow;
			grow.ino = gno1;
			gmatrix.push_back(grow);

			gmindex.insert(IndexVal(gno1, growno));
		}
		else growno = grit->second;				// use existing gene row


		// loop over in-row proteins
		for (int jp = 0; jp < (int)prow.cell.size(); jp++)
		{
			Cell & pcell = prow.cell[jp];
			Protein & prot2 = proteins[pcell.jno];
			int gno2 = prot2.gno;

			// add/modify gmatrix[gno1][gno2]
			Row & grow = gmatrix[growno];
			int jg = 0;
			for ( ; jg < (int)grow.cell.size(); jg++)
			{
				Cell & gcell = grow.cell[jg];
				if (gcell.jno == gno2)
				{
					if (pcell.score > gcell.score) gcell.score = pcell.score;
					if (pcell.eval < gcell.eval) gcell.eval = pcell.eval;
					break;
				}		// gmatrix cell modified
			}		// next gene over grow

			if (jg >= (int)grow.cell.size())	// add cell
			{
				Cell gcell(pcell);
				gcell.jno = gno2;
				grow.cell.push_back(gcell);
			}		// gmatrix cell added
		}		// next protein over prow
	}		// next row of pmatrix

	// calculate nrows & ncells
	int nrows = 0, ncells = 0;
	for (int i = 0; i < (int)gmatrix.size(); i++)
	{
		Row & row = gmatrix[i];
		if (!row.cell.empty())
		{
			nrows++;
			ncells += (int)row.cell.size();
		}
	}

	ostringstream oss;
    if (mpisize > 1) oss << "[" << right << setw(rankw) << myrank << "] "; 
    oss << spec.name;
	string pref = oss.str();
	oss.str(""); oss.clear();
	oss << left << setw(maxw) << pref << right << " gen=" << setw(5)
		<< nrows << "  cell=" << setw(7) << ncells;
	cerr << oss.str() << endl;
	if (det.is_open()) det << oss.str() << endl;
}


// sort and output the matrix to the result file
// aggregate: 0=pmatrix  1=gmatrix
void	outputResult (const int sp)
{
	int nrows = 0, ncells = 0;
	ostringstream oss;
	Species & spec = species[sp];
	vector <Row> & matrix = (aggregate == 0 ? pmatrix : gmatrix);
    if (mpisize > 1) oss << "[" << right << setw(rankw) << myrank << "] "; 
    oss << spec.name;

	// sort the matrix if needed
	if (!vsortmode.empty() || !hsortmode.empty())
	{
		if (!hsortmode.empty())		// sort inside each row
			for (int i = 0; i < (int)matrix.size(); i++)
			{
				Row & row = matrix[i];
                current_ino = row.ino;
				sort(row.cell.begin(), row.cell.end(), sortCells);
			}

		if (!vsortmode.empty())		// sort rows of the matrix
			sort(matrix.begin(), matrix.end(), sortRows);

		oss << " sorted and";
	}

	string filename = outpath + spec.name + outext;
	ofstream out;
	out.open(filename.c_str(), ios_base::binary);
	if (!out.good())
	{
		oss << " open failed";
		cerr << oss.str() << endl;
		if (det.is_open()) det << oss.str() << endl;
	#ifndef NOMPI
		if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_ERESULT);
	#endif
		exit(RETCODE_ERESULT);
	}

	// loop over the matrix rows
	for (int i = 0; i < (int)matrix.size(); i++)
	{
		Row & row = matrix[i];
		int ino = row.ino;
		string id1 = (aggregate == 0 ? proteins[ino].id : genes[ino].id);
		bool are_cells = false;

		vector <int> perspp(nsp, 0);
        vector <float> maxscore(nsp, 0.0);

		for (int j = 0; j < (int)row.cell.size(); j++)
		{
			Cell & cell = row.cell[j];
			int jno = cell.jno;
			string id2 = aggregate == 0 ? proteins[jno].id : genes[jno].id;
			int    sp2 = aggregate == 0 ? proteins[jno].sp : genes[jno].sp;
			perspp[sp2] ++;
            if (ratio != 0 && maxscore[sp2] == 0) maxscore[sp2] = cell.score;

			if (maxhomo != 0 && perspp[sp2] > maxhomo) continue;
            if (ratio != 0 && 1.0 - cell.score / maxscore[sp2] > ratio) continue;

			out << id1 << tab << id2 << tab /*<< fixed*/
				<< setprecision(useweight == 0 ? 0 : 3) << cell.score << tab
				<< /*scientific <<*/ setprecision(2) << cell.eval /*<< fixed*/;

            if (addrank != 0)	// add rank field
            {
                float outrank = ratio == 0 ? (float)perspp[sp2] : 1.0F - cell.score / maxscore[sp2];
                out << tab << /*fixed << setprecision((int)(ratio==0?0:3)) <<*/ setprecision(3) << outrank;

                if (addrank > 1)            // store rank in backward matrix
                {
                    Brank brank;
                    brank.ino = ino;
                    brank.rank = outrank;

                    // try add new column
                    int ncolumn = (int)rmatrix.size();
                    pair <IndexIt, bool> res;
                    res = rmindex.insert(IndexVal(jno, ncolumn));

                    if (!res.second)        // column exists
                    {
                        int jcol = res.first->second;       // column no.
                        Column & column = rmatrix[jcol];
                        column.brank.push_back(brank);
                    }
                    else                    // add new column
                    {
                        Column column;
                        column.jno = jno;
                        column.brank.push_back(brank);
                        rmatrix.push_back(column);
                    }
                }
            }
			out << endl;
			are_cells = true;
			ncells++;
		}
		if (are_cells) nrows++;
	}

	gene_rows += nrows;
	gene_cells += ncells;

	out.close();
	oss << " written to file: row=" << nrows << "  cell=" << ncells;
	cerr << oss.str() << endl;
	if (det.is_open()) det << oss.str() << endl;
}

// read result file written by another process
bool	readResult(const int sp)
{
    ostringstream oss;
    Species & spec = species[sp];
    string filename = outpath + spec.name + outext;

    if (verbose != 0) {
        if (mpisize > 1) oss << "[" << right << setw(rankw) << myrank << "] ";
        oss << left << setw(maxw) << spec.name << ":";
    }

    ifstream in;       // read back out
    in.open(filename.c_str(), ios_base::binary);
    if (!in.good())
    {
        oss << " open failed";
        cerr << oss.str() << endl;
        if (det.is_open()) det << oss.str() << endl;
#ifndef NOMPI
        if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_ERESULT);
#endif
        exit(RETCODE_ERESULT);
    }

    string line;
    int lineno = 0;
    int nerr = 0;

    while (!in.eof())      // file read loop
    {
        in.getline(buf, buflen);
        if (!in.good()) break;
        line.assign(buf);
        if (line.empty()) break;
        lineno++;

        string id1, id2;
        double score, evalue;
        float rank;
        istringstream iss(line);
        iss >> id1 >> id2 >> score >> evalue >> rank;
        if (iss.fail())
        {
            if (myrank == 0)
            {
                oss << " error in line " << lineno;
                cerr << oss.str() << endl;
                if (det.is_open()) det << oss.str() << endl;
            }
            return false;
        }

        int ino, jno;
        if (aggregate == 0)
        {
            DictIt dit = prodict.find(id1);
            ino = dit == prodict.end() ? -1 : dit->second;
            dit = prodict.find(id2);
            jno = dit == prodict.end() ? -1 : dit->second;
        }
        else
        {
            DictIt dit = gendict.find(id1);
            ino = dit == gendict.end() ? -1 : dit->second;
            dit = gendict.find(id2);
            jno = dit == gendict.end() ? -1 : dit->second;
        }

        if (ino < 0 || jno < 0) { nerr++; continue; }

        Brank brank;
        brank.ino = ino;
        brank.rank = rank;

        // try add new column
        int ncolumn = (int)rmatrix.size();
        pair <IndexIt, bool> res;
        res = rmindex.insert(IndexVal(jno, ncolumn));

        if (!res.second)        // column exists
        {
            int jcol = res.first->second;       // column no.
            Column & column = rmatrix[jcol];
            column.brank.push_back(brank);
        }
        else                    // add new column
        {
            Column column;
            column.jno = jno;
            column.brank.push_back(brank);
        }
    }
    in.close();

    if (verbose != 0 && myrank == 0)
    {
        oss << " lines=" << lineno;
        if (nerr > 0) oss << " err=" << nerr;
        cerr << oss.str() << endl;
        if (det.is_open()) det << oss.str() << endl;
    }
    return true;
}

// read existing result file, write one appended with 2nd rank and swap
bool    writeBackRank(const int sp)
{
    ostringstream oss;
    Species & spec = species[sp];
    string filename = outpath + spec.name + outext;
    string filename2 = outpath + spec.name + ".new";

    if (mpisize > 1) oss << "[" << right << setw(rankw) << myrank << "] "; 
    oss << left << setw(maxw) << spec.name << ":";

    ifstream in;       // read back
    ofstream out;
    in.open(filename.c_str(), ios_base::binary);
    out.open(filename2.c_str(), ios_base::binary);
    if (!in.good() || !out.good())
    {
        oss << " open failed";
        cerr << oss.str() << endl;
        if (det.is_open()) det << oss.str() << endl;
#ifndef NOMPI
        if (mpisize > 1) MPI_Abort(MPI_COMM_WORLD, RETCODE_ERESULT);
#endif
        exit(RETCODE_ERESULT);
    }

    string line;
    int lineno = 0;
    int nerr = 0;       // number of erroneous lines
    int nnor = 0;       // number of noranks

    while (!in.eof()) 
    {
        in.getline(buf, buflen);
        if (!in.good()) break;
        line.assign(buf);
        if (line.empty()) break;
        lineno++;

        string id1, id2;
        double score, evalue;
        int rank;
        istringstream iss(line);
        iss >> id1 >> id2 >> score >> evalue >> rank;
        if (iss.fail())
        {
            if (myrank == 0)
            {
                oss << " error in line " << lineno;
                cerr << oss.str() << endl;
                if (det.is_open()) det << oss.str() << endl;
            }
            return false;
        }

        int ino, jno;
        if (aggregate == 0)
        {
            DictIt dit = prodict.find(id1);
            ino = dit == prodict.end() ? -1 : dit->second;
            dit = prodict.find(id2);
            jno = dit == prodict.end() ? -1 : dit->second;
        }
        else
        {
            DictIt dit = gendict.find(id1);
            ino = dit == gendict.end() ? -1 : dit->second;
            dit = gendict.find(id2);
            jno = dit == gendict.end() ? -1 : dit->second;
        }

        if (ino < 0 || jno < 0) 
        { 
            nerr++;
            out << line << endl;
            continue; 
        }

        IndexIt it = rmindex.find(ino);     // NB: inverted
        if (it == rmindex.end())
        {
            nnor++;
            out << line << tab << norank << endl;
            continue; 
        }

        Column & column = rmatrix[it->second];
        int k = 0;
        for ( ; k < (int)column.brank.size(); k++)
            if (column.brank[k].ino == jno) break;     // NB: inverted

        if (k < (int)column.brank.size())
            out << line << tab /*<< fixed << setprecision((int)(ratio==0?0:3))*/ 
                << setprecision(3) << column.brank[k].rank << endl;
        else
        {
            if (ratio == 0)
                out << line << tab << norank << endl;
            else
                out << line << tab << /*fixed << setprecision(1) <<*/ 1.0 << endl;
            nnor++;
        }
    }
    in.close();
    out.close();
    remove(filename.c_str());                           // just replace
    rename(filename2.c_str(), filename.c_str());

    oss << " lines=" << lineno << " norank=" << nnor;
    if (nerr > 0) oss << " err=" << nerr;
    cerr << oss.str() << endl;
    if (det.is_open()) det << oss.str() << endl;
    return true;
}

// sort cells of the matrix row
bool	sortCells (const Cell& c1, const Cell& c2)
{
	for (int k = 0; k < (int)hsortmode.size(); k++)
	{
		char c = hsortmode[k];

		if (c == 's' || c == 'S')
		{
			int s1 = aggregate == 0 ? proteins[c1.jno].sp : genes[c1.jno].sp;
			int s2 = aggregate == 0 ? proteins[c2.jno].sp : genes[c2.jno].sp;
			if (s1 < s2) return (c == 's');
			if (s1 > s2) return (c == 'S');
		}

        else if (c == '1')
        {
            if (aggregate == 0) {
                string id0 = proteins[current_ino].id;
                string id1 = proteins[c1.jno].id;
                if (id0 == id1) return true;
                string id2 = proteins[c2.jno].id;
                if (id0 == id2) return false;
            }
            else {
                int i0 = atoi(genes[current_ino].id.c_str());
                int i1 = atoi(genes[c1.jno].id.c_str());
                if (i0 == i1) return true;
                int i2 = atoi(genes[c2.jno].id.c_str());
                if (i0 == i2) return false;
            }
        }

		else if (c == 'i' || c == 'I')
		{
            if (aggregate == 0) {
                string id1 = proteins[c1.jno].id;
                string id2 = proteins[c2.jno].id;
                if (id1 < id2) return (c == 'i');
                if (id1 > id2) return (c == 'I');
            }
            else {
                int i1 = atoi(genes[c1.jno].id.c_str());
                int i2 = atoi(genes[c2.jno].id.c_str());
                if (i1 < i2) return (c == 'i');
                if (i1 > i2) return (c == 'i');
            }
		}

		else if (c == 'w' || c == 'W')
		{
			if (c1.score < c2.score) return (c == 'w');
			if (c1.score > c2.score) return (c == 'W');
		}

		else if (c == 'e' || c == 'E')
		{
			if (c1.eval < c2.eval) return (c == 'e');
			if (c1.eval > c2.eval) return (c == 'E');
		}

		else if (c == 'p' || c == 'P')
		{
			int g1 = aggregate == 0 ? proteins[c1.jno].gno : c1.jno;
			int g2 = aggregate == 0 ? proteins[c2.jno].gno : c2.jno;
			if (genes[g1].region < genes[g2].region) return (c == 'p');
			if (genes[g1].region > genes[g2].region) return (c == 'P');
			if (genes[g1].start < genes[g2].start) return (c == 'p');
			if (genes[g1].start > genes[g2].start) return (c == 'P');
		}
	}
	return true;
}

// sort rows of the matrix
bool	sortRows (const Row& r1, const Row& r2)
{
	for (int k = 0; k < (int)vsortmode.size(); k++)
	{
		char c = vsortmode[k];
		if (c == 's' || c == 'S')
		{
			int s1 = aggregate == 0 ? proteins[r1.ino].sp : genes[r1.ino].sp;
			int s2 = aggregate == 0 ? proteins[r2.ino].sp : genes[r2.ino].sp;
			if (s1 < s2) return (c == 's');
			if (s1 > s2) return (c == 'S');
		}

		else if (c == 'i' || c == 'I')
		{
			string id1 = aggregate == 0 ? proteins[r1.ino].id : genes[r1.ino].id;
			string id2 = aggregate == 0 ? proteins[r2.ino].id : genes[r2.ino].id;
			if (id1 < id2) return (c == 'i');
			if (id1 > id2) return (c == 'I');
		}

		else if (c == 'p' || c == 'P')
		{
			int g1 = aggregate == 0 ? proteins[r1.ino].gno : r1.ino;
			int g2 = aggregate == 0 ? proteins[r2.ino].gno : r2.ino;
			if (genes[g1].region < genes[g2].region) return (c == 'p');
			if (genes[g1].region > genes[g2].region) return (c == 'P');
			if (genes[g1].start < genes[g2].start) return (c == 'p');
			if (genes[g1].start > genes[g2].start) return (c == 'P');
		}
	}
	return true;
}
